/** @file dumpw.h

    written by Marc Singer
    4 Jun 2008

    Copyright (C) 2008 Marc Singer

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.  Please
    refer to the file debian/copyright for further details.

*/

#if !defined (__DUMPW_H__)
#    define   __DUMPW_H__

/* ----- Includes */

/* ----- Types */

/* ----- Globals */

/* ----- Prototypes */

void dumpw (const char *rgb, int cb, unsigned long index, int width);


#endif  /* __DUMPW_H__ */
